/* Hard disk drive temperature
 * Copyright (C) 2001  Emmanuel VARAGNAT <coredump@free.fr>

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <errno.h>
#include <regex.h>
#include <sys/ioctl.h>
#include <linux/hdreg.h>

#include "hddtemp.h"

#ifdef ARCH_I386
typedef unsigned short u16;
#endif

#define swapb(x) \
({ \
        u16 __x = (x); \
        (x) = ((u16)( \
                (((u16)(__x) & (u16)0x00ffU) << 8) | \
                (((u16)(__x) & (u16)0xff00U) >> 8) )); \
})
#define F_to_C(val) (int)(((double)(val)-32.0)/1.8)
#define C_to_F(val) (int)(((double)(val)*1.8)+32)

#define HDDTEMP_VERSION        "0.2"
#define DEFAULT_ATTRIBUTE_ID   194
#define DEFAULT_DATABASE_PATH  "/usr/share/hddtemp.db"
#define MAX_LINE_LEN           1024

struct harddrive_entry   *supported_drives = NULL;
struct harddrive_entry   **last_entry = &supported_drives;
struct harddrive_entry   *drive_entry;
int                      debug, quiet;
char                     *database_path = DEFAULT_DATABASE_PATH;

struct harddrive_entry *is_a_supported_drive(const char* model) {
  struct harddrive_entry *p;

  for(p = supported_drives; p; p = p->next) {
    regex_t     preg;
    regmatch_t  pmatch;

    if(regcomp(&preg, p->regexp, 0))
      exit(-2);
    if(regexec(&preg, model, 1, &pmatch, 0) == 0)
      return p;
  }

  return NULL;
}

const char *extract_string(char **string) {
  char *str;

  if(**string != '"')
    return NULL;
  else
    (*string)++;

  str = *string;

  for(; *str; str++) {
    switch(*str) {
      case '"':
	 return str; 
         break;
      case '\\':
         str++;
         break;
    }
  }

  return NULL;
}

int parse_db_line(char *line) {
  const char             *regexp, 
                         *description;
  struct harddrive_entry *new_entry;
  unsigned short         value;
  char                   unit;
  char                   *p;

  line += strspn(line, " \t");
  if(*line == '#' || *line == '\0')
    return 0;

  /* extract regexp */
  p = (char*) extract_string(&line); 
  if(p == NULL || *p == '\0')
    return 1;

  *p = '\0';
  regexp = line;
  line = p + 1;

  /* extract value */
  line += strspn(line, " \t");
  if(! *line)
    return 1;

  p = line;
  value = 0;
  while(*p >= '0' && *p <= '9') {
    value = 10*value + *p-'0';
    p++;
  }
  if(*p == '\0' || ( *p != ' ' && *p != '\t') )
    return 1;
  line = p;

  /* extract temperature units */
  line += strspn(line, " \t");
  if(! *line)
    return 1;

  if((*line != 'C' && *line != 'F') || ( *(line+1) != ' ' && *(line+1) != '\t' ))
    return 1;

  unit = *(line++);

  /* extract description */
  line += strspn(line, " \t");
  if(! *line)
    return 1;

  p = (char*) extract_string(&line);
  if(p == NULL)
    return 1;

  description = line; 
  *p = '\0';

  new_entry = (struct harddrive_entry*) malloc(sizeof(struct harddrive_entry));
  if(new_entry == NULL) {
    perror("malloc");
    exit(-1);
  }
  
  new_entry->regexp = strdup(regexp);
  new_entry->description = strdup(description);
  new_entry->attribute_id = value;
  new_entry->unit = unit;
  new_entry->next = NULL;
  *last_entry = new_entry;
  last_entry = &(new_entry->next);

  return 0;
}

void load_database() {
  char  buff[MAX_LINE_LEN];
  char  *p, *s, *e, *ee;
  int   fd;
  int   n, rest, numline;

  errno = 0;
  fd = open(database_path, O_RDONLY);
  if(fd == -1) {
    fprintf(stderr, "hddtemp: Can't open %s: %s\n", database_path, strerror(errno));
    exit(2);
  }

  numline = 0;
  rest = MAX_LINE_LEN;
  p = buff;
  while((n = read(fd, p, rest)) > 0) {
    s = buff;
    ee = p + n;
    while((e = memchr(s, '\n', ee-s))) {
      *e = '\0';
      numline++;
      if(parse_db_line(s)) {
	fprintf(stderr, "ERROR: syntax error at line %d in %s\n", numline, database_path);
        exit(2);
      }
      s = e+1;
    }

    if(s == buff) {
      fprintf(stderr, "  ERROR: a line exceed %d characters in %s.\n", MAX_LINE_LEN, database_path);
      close(fd);
      exit(2);
    }

    memmove(buff, s, ee-s); /* because memory can overlap */

    rest = MAX_LINE_LEN - (ee-s);
    p = buff + (ee-s);
  }

  close(fd);
}

void show_supported_drives() {
  unsigned char           *tabs, *line;
  struct harddrive_entry  *p;
  int                     max, len;

  max = 0;
  for(p = supported_drives; p; p = p->next) {
    len = strlen(p->regexp);
    if(len > max)
      max = len;
  }
  len = max/8 + 1;
  tabs = malloc(len+1);
  memset(tabs, '\t', len);
  tabs[len] = '\0';

  len = (max/8 + 1) * 8;
  line = malloc(len+1);
  memset(line, '-', len);
  line[len] = '\0';
  
  printf("\n"
	 "Regexp%s| Value | Description\n"
	 "------%s---------------------\n", tabs, line);

  for(p = supported_drives; p; p = p->next) {
    len = strlen(p->regexp);
    printf("%s%s| %5d | %s\n",
	   p->regexp,
	   tabs+(len/8),
	   p->attribute_id,
	   p->description);
  }  
  printf("\n");
}


int enable_smart(int fd) {
  unsigned char cmd[4] = { WIN_SMART, 0, SMART_ENABLE, 0 };

  return ioctl(fd, HDIO_DRIVE_CMD, cmd);
}


int get_smart_values(int fd, unsigned char* buff) {
  unsigned char  cmd[516] = { WIN_SMART, 0, SMART_READ_VALUES, 1 };
  int            ret;

  ret = ioctl(fd, HDIO_DRIVE_CMD, cmd);
  if(ret)
    return ret;

  memcpy(buff, cmd+4, 512);
  return 0;
}

/*
int get_smart_threshold_values(int fd, unsigned char* buff) {
  unsigned char cmd[516] = { WIN_SMART, 0, SMART_READ_THRESHOLDS, 1 };
  int ret;

  ret = ioctl(fd, HDIO_DRIVE_CMD, cmd);
  if(ret)
    return ret;

  memcpy(buff, cmd+4, 512);
  return 0;
}
*/

unsigned char* search_temperature(const unsigned char* smart_data, int attribute_id) {
  int i, n;

  n = 3;
  i = 0;
  while((debug || *(smart_data + n) != attribute_id) && i < 30) {
    if(debug && *(smart_data + n))
      printf("field(%d)\t = %d\n", *(smart_data + n), *(smart_data + n + 3));

    n += 12;
    i++;
  }

  if(i >= 30)
    return NULL;
  else
    return (unsigned char*)(smart_data + n);
}


void display_temperature(char *drive) {
  unsigned char identity[512];
  unsigned char values[512]/*, thresholds[512]*/;
  unsigned char *field;
  char          *model;
  int           fd;
  int           i;
  u16           *p;

  errno = 0;
  fd = open(drive, O_RDONLY);
  if(fd < 0) {
    fprintf(stderr, "%s: open: %s\n", drive, strerror(errno));
    return;
  }

  /* Get identity drive for model name */
  if(ioctl(fd, HDIO_GET_IDENTITY, identity))
    model = "?????";
  else
    model = (char*) ((u16*)identity + 27);

  drive_entry = is_a_supported_drive(model);

  if(drive_entry && drive_entry->attribute_id == 0) {
    printf("%s: %s:  known (found in database) to do not have temperature sensor.\n", drive, model);
    close(fd);
    return;
  }

  /* get SMART values */
  if(enable_smart(fd) != 0) {
    printf("%s: %s:  S.M.A.R.T. not available\n", drive, model);
    close(fd);
    return;
  }

  if(get_smart_values(fd, values)) {
    perror("ioctl");
    close(fd);
    return;
  }

  close(fd);

  p = (u16*)values;
  for(i = 0; i < 256; i++) {
    swapb(*(p+i));
  }

  /* get SMART threshold values */
  /*
  if(get_smart_threshold_values(fd, thresholds)) {
    perror("ioctl");
    exit(3);
  }

  p = (u16*)thresholds;
  for(i = 0; i < 256; i++) {
    swapb(*(p+i));
  }
  */

  /* temperature */
  if(drive_entry && drive_entry->attribute_id > 0)
    field = search_temperature(values, drive_entry->attribute_id);
  else
    field = search_temperature(values, DEFAULT_ATTRIBUTE_ID);

  if(debug) {
    printf("\n Model: %s\n\n", model);
    return;
  }

  if(drive_entry && field)
    printf("%s: %s:  %dC\n",
           drive,
           model,
           (drive_entry->unit == 'C') ? *(field+3) : F_to_C(*(field+3)));
  else {
    if(field) {
      if(!quiet)
	printf("WARNING: Drive %s doesn't appear in the database of supported drives\n"
	       "WARNING: But using a common value, it reports something.\n"
	       "WARNING: Note that the temperature shown could be wrong.\n"
	       "WARNING: See --help, --debug and --drives options.\n"
	       "WARNING: And don't forget you can add your drive to hddtemp.db\n", drive);
      printf("%s: %s:  %dC or F\n", drive, model, *(field+3));
    }
    else {
      if(!quiet)
	printf("WARNING: Drive %s doesn't seem to have a temperature sensor.\n"
	       "WARNING: This doesn't mean it hasn't got one.\n"
	       "WARNING: If you are sure it has one, please contact me (coredump@free.fr).\n"
	       "WARNING: See --help, --debug and --drives options.\n", drive);
      printf("%s: %s:  no sensor\n", drive, model);
    }
  }

  /* threshold temperature */
  /*
  field = search_temperature(thresholds);
  if(field)
    printf("%s: %s  %dC\n", drive, model, *(field+1));
  else
    printf("Your drive doesn't seem to have temperature sensor.\n"
	   "If you are sure it has one, please contact me (coredump@free.fr).\n");
  */
}


int main(int argc, char* argv[]) {
  int i, c, lindex = 0;
  int show_db;

  show_db = debug = quiet = 0;

  /* Parse command line */
  optind = 0;
  while (1) {
    static struct option long_options[] = {
      {"help",    0, NULL, 'h'},
      {"quiet",   0, NULL, 'q'},
      {"drives",  0, NULL, 'd'},
      {"file",    1, NULL, 'f'},
      {"debug",   0, NULL, 'D'},
      {"version", 0, NULL, 'v'},
      {0, 0, 0, 0}
    };
 
    c = getopt_long (argc, argv, "Ddf:hqv", long_options, &lindex);
    if (c == -1)
      break;
    
    switch (c) {
      case 'q':
	quiet = 1;
	break;
      case 'd':
	show_db = 1;
	break;
      case 'D':
	debug = 1;
	break;
      case 'f':
	database_path = optarg;
	break;
      case '?':
      case 'h':
	printf("Usage: hddtemp [OPTION] DISK1 [DISK2]...\n\n"
	       "   hddtemp display the temperature of the IDE drive specified in argument.\n"
	       "   Your drive must support S.M.A.R.T. capabilities.\n"
	       "   For the moment only few drives are supported.\n"
	       "\n"
	       "  -D    --debug   :  display various S.M.A.R.T. fields and their values.\n"
	       "                     Useful to find a value that seems to match the temperature\n"
	       "                     and/or to send me a report.\n"
	       "                     (only done for the first drive supplied).\n"
	       "  -d    --drives  :  show regular expressions that allow hddtemp to recognize\n"
	       "                     a supported drive.\n"
	       "  -f    --file    :  specify the database file.\n"
	       "  -q    --quiet   :  doesn't check if the drive drive is supported.\n"
	       "  -v    --version :  display hddtemp version number.\n"
	       "\n");
      case 'v':
	printf("hddtemp version %s\n", HDDTEMP_VERSION);
	exit(0);
	break;
      default:
	exit(1);
      }
  }
  
  if(show_db) {
     load_database();
     show_supported_drives();
     exit(0);
  }
  
  if(geteuid()) {
    fprintf(stderr,
	    "  ERROR: You must be root to run the command,\n"
	    "  ERROR: or the root must set the suid bit for the executable.\n");
    exit(1);
  }

  if(argc - optind <= 0) {
    fprintf(stderr, "Too few arguments: you must specify a drive.\n");
    exit(1);
  }

  if(debug) {
    argc = optind + 1;
    quiet = 1;
  }

  load_database();

  /* Open devices */
  for(i = optind; i < argc; i++) {
    display_temperature(argv[i]);
  }

  if(debug) {
    printf("If one of the field value seems to match the temperature, you can\n"
	   "send me a report so I can add it to the database of supported drives\n"
	   "and/or you can do it yourself by adding an entry in hddtemp.db and send\n"
	   "me the updated hddtemp.db.\n"
	   "\n"
	   "Don't forget to compare your value(s) with those in the database.\n"
	   "(see option -d)\n");
  }

  return 0;
}
